"""Retrieve data from external APIs."""
from __future__ import annotations

import functools
import json
from enum import Enum

import requests


class EndPoints(Enum):

    """Endpoints we can retrieve data from."""

    MUNICIPALITIES = "https://ws.geonorge.no/kommuneinfo/v1/kommuner"


@functools.cache
def get_municipalities() -> list[dict[str, str]]:
    """Get a list of municipalities in Norway."""
    response: requests.Response = requests.get(
        EndPoints.MUNICIPALITIES.value,
        timeout=5,
    )
    return json.loads(response.text)



"""Get and generate synthetic data from a variety of sources."""
from __future__ import annotations

import random
import typing
from typing import TYPE_CHECKING

from datadoc_model import Enums
from faker import Faker

if TYPE_CHECKING:
    from datetime import date

faker = Faker("no_NO")


def is_ssid(short_name: str) -> bool:
    """Return True if this column is a Norwegian Personal Number."""
    short_name = short_name.lower().strip().replace("_", "")
    return "fnr" in short_name or "snr" in short_name


FUNCTION_COLUMN_NAME_MAPPING = {
    faker.name: ["navn"],
    faker.iana_id: ["kod"],
    faker.city: ["skolenavn", "sted", "by"],
    faker.date: ["dato", "start", "avsl"],
}


def get_string(
    column_name: str,
    mapping: dict[typing.Callable[[], str], list[str]] = FUNCTION_COLUMN_NAME_MAPPING,
) -> str:
    """Generate synthetic string data."""
    column_name = column_name.lower()

    if is_ssid(column_name):
        # TODO @mmwinther: Retrieve fnr/snr from our shared test catalog
        # https://github.com/statisticsnorway/metamock/issues/14
        return faker.ssn()

    for function, name_parts in mapping.items():
        if any(bool(part in column_name.lower().strip()) for part in name_parts):
            return function()

    return faker.word()


def get_datetime() -> date:
    """Generate synthetic date data."""
    return faker.date_object()


def get_integer(
    column_name: str,
    row_number: int,
    *,
    identifier: bool = False,
) -> int:
    """Generate synthetic integer data."""
    column_name = column_name.lower()

    match column_name:
        case "alder":
            return faker.random_int(min=0, max=110)

        case "hkltrinn":
            return faker.random_int(min=1, max=13)

    if identifier is True:
        return faker.unique.random_int(min=0, max=row_number * 2)

    return random.randint(0, 1000)  # noqa:S311 non-cryptographic


def generate_data_for_data_type(
    data_type: Enums.Datatype,
    column_name: str,
    length: int,
    *,
    identifier: bool = False,
) -> list[typing.Any] | None:
    """Generate synthetic data for the given data type."""

    def generate_string(column_name: str = column_name) -> str:
        """Pipe arguments to the get_string function."""
        return get_string(column_name)

    def generate_integer(
        column_name: str = column_name,
        length: int = length,
        *,
        identifier: bool = identifier,
    ) -> int:
        """Pipe the arguments to the get_integer function."""
        return get_integer(column_name, length, identifier=identifier)

    functions: dict[Enums.Datatype, typing.Callable[[], typing.Any]] = {
        Enums.Datatype.STRING: generate_string,
        Enums.Datatype.INTEGER: generate_integer,
        Enums.Datatype.DATETIME: lambda: get_datetime(),
        Enums.Datatype.BOOLEAN: lambda: random.choice(  # noqa: S311 non-cryptographic
            [True, False],
        ),
        Enums.Datatype.FLOAT: lambda: random.uniform(  # noqa: S311 non-cryptographic
            0,
            1000,
        ),
    }
    try:
        return [functions[data_type]() for _ in range(length)]
    except KeyError:
        return None



"""Functionality relating to Klass."""

from __future__ import annotations

import random
from typing import TYPE_CHECKING

from klass.classes.classification import KlassClassification

if TYPE_CHECKING:
    import pandas as pd

# Keep this dict in numeric order
CLASSIFICATION_COLUMN_NAME_MAPPING: dict[str, list[str]] = {
    "1": ["fskolenr"],
    "2": ["kjonn", "kjoenn"],
    "6": ["nace", "naering"],
    "31": ["mrk_dl"],
    "36": ["nus2000", "emnekode", "vidutd", "utdkode"],
    "66": ["fakultet"],
    "91": ["stat", "borg"],
    "100": ["land"],
    "127": ["fylk"],
    "131": ["skolekom", "komm"],
}

KLASS_URI_BASE: str = "https://www.ssb.no/klass/klassifikasjoner/"


def get_codes_for_klass_classification(classification_id: str) -> pd.Series:
    """Get codes for a given Classification ID."""
    return KlassClassification(classification_id).get_codes().data["code"]


def generate_data_from_klass_codes(
    column_name: str | None,
    classification_uri: str | None,
    length: int,
) -> list | None:
    """Generate a column of synthetic data based on the code list from a classification.

    If there is a defined URI in the metadata, use that.
    If we can guess a classification based on the name of the column, use that.
    If we can't find a matching classification, just return None.
    """
    classification_id = None
    if classification_uri is not None and classification_uri.startswith(
        KLASS_URI_BASE,
    ):
        # First choice is to use the classification URI defined in the metadata
        classification_id = classification_uri.replace(
            KLASS_URI_BASE,
            "",
        )
    else:
        classification_id = guess_classification_from_name(
            column_name,
        )

    if not classification_id:
        # We didn't find a matching classification so stop here
        return None

    return [
        random.choice(  # noqa:S311 non-cryptographic
            get_codes_for_klass_classification(
                classification_id,  # type: ignore [arg-type]
            ),
        )
        for _ in range(length)
    ]


def is_code(column_name: str | None) -> bool:
    """Return true if we should use the classification code, otherwise we use the value."""
    if not column_name:
        return False
    return "kode" in column_name or "nr" in column_name


def guess_classification_from_name(
    column_name: str | None,
    mapping: dict[str, list[str]] = CLASSIFICATION_COLUMN_NAME_MAPPING,
) -> str | None:
    """Get a classification ID if we know of one."""
    if not column_name:
        return None
    for classification_id, name_parts in mapping.items():
        if any(bool(part in column_name.lower().strip()) for part in name_parts):
            return classification_id
    return None



"""Generate synthetic data from a Datadoc document."""

from __future__ import annotations

import argparse
import json
import logging
import pathlib
import sys
from enum import Enum
from pprint import pprint

import numpy as np
import pandas as pd
from datadoc_model import Enums
from datadoc_model.Model import DataDocVariable

from metamock.faker_data import (
    generate_data_for_data_type,
)
from metamock.klass_classifications_data import (
    generate_data_from_klass_codes,
)

DEFAULT_NUMBER_OF_ROWS = 10
MAX_ROWS = 1_000_000


class FileType(Enum):

    """Supported file types."""

    CSV = "csv"
    PARQUET = "parquet"
    JSON = "json"


logger = logging.getLogger(__name__)


def read_metadata_file(path: pathlib.Path) -> list[DataDocVariable]:
    """Open the given file and read the metadata into Datadoc models."""
    fresh_metadata = {}

    with path.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)

    variables_list = fresh_metadata.pop("variables", None)
    return [DataDocVariable(**v) for v in variables_list]


def generate_synthetic_data(
    variable_metadata: list[DataDocVariable],
    rows: int,
) -> pd.DataFrame:
    """Generate synthetic data from the given list of DataDocVariable."""
    # Fill the generated data with NaN as a starting point
    generated_data: pd.DataFrame = pd.DataFrame(
        np.empty((rows, len(variable_metadata))) * np.nan,
        columns=[v.short_name for v in variable_metadata],
    )

    for v in variable_metadata:
        data_list = None
        identifier: bool = v.variable_role == Enums.VariableRole.IDENTIFIER

        # First, attempt to get data from Klass since this is the most relevant for SSB.
        data_list = generate_data_from_klass_codes(
            v.short_name,
            v.classification_uri,
            rows,
        )

        # We couldn't find anything in Klass, so we fall back to generating from the data type
        if not data_list and v.data_type:
            data_list = generate_data_for_data_type(
                Enums.Datatype(v.data_type),
                v.short_name,
                rows,
                identifier=identifier,
            )

        if data_list:
            generated_data[v.short_name] = data_list

    return generated_data


def save_data_to_file(
    data: pd.DataFrame,
    folder_path: pathlib.Path,
    file_name: str,
    file_type: FileType,
) -> None:
    """Save the given dataframe to the specified file path."""
    file_name = (
        file_name.replace(FileType.CSV.value, "")
        .replace(FileType.JSON.value, "")
        .replace(FileType.PARQUET.value, "")
    )
    output_path = folder_path / f"{file_name}.{file_type.value.lower()}"
    match file_type:
        case FileType.CSV:
            data.to_csv(
                output_path,
                sep=";",
                encoding="utf-8",
            )

        case FileType.PARQUET:
            data.to_parquet(output_path)

        case FileType.JSON:
            data.to_json(output_path)

    print(f"Successfully generated synthetic data: {output_path}")  # noqa: T201


def main(sys_args: list[str] = sys.argv[1:]) -> None:
    """Run the app."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--datadoc-document", type=pathlib.Path)
    parser.add_argument("-f", "--output-path", type=pathlib.Path)
    parser.add_argument("-n", "--output-filename", type=str)
    parser.add_argument(
        "-t",
        "--file-type",
        type=str,
        default="parquet",
        choices=[v.value for v in FileType],
    )
    parser.add_argument("-r", "--no-of-rows", type=int, default=DEFAULT_NUMBER_OF_ROWS)
    args = parser.parse_args(sys_args)
    file_type = FileType(args.file_type)

    if args.no_of_rows < 1 or args.no_of_rows > MAX_ROWS:
        parser.error(f"Number of rows must be between 1 and {MAX_ROWS}")

    variable_metadata = read_metadata_file(args.datadoc_document)

    if insufficient_metadata := [
        v.short_name for v in variable_metadata if not all([v.data_type, v.short_name])
    ]:
        print(  # noqa: T201
            "WARNING: The following columns contain insufficient metadata and will have empty values in the output dataset.",
        )
        pprint(  # noqa: T203
            insufficient_metadata,
        )

    generated_data: pd.DataFrame = generate_synthetic_data(
        variable_metadata,
        args.no_of_rows,
    )

    save_data_to_file(
        generated_data,
        args.output_path,
        args.output_filename,
        file_type,
    )


if __name__ == "__main__":
    main()
